
import json
from urllib.request import urlretrieve
import os
import os.path as osp
import zipfile

import numpy as np
from tqdm import tqdm

from airfrans.simulation import Simulation

class DownloadProgressBar(tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`."""
    def update_to(self, b = 1, bsize = 1, tsize = None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b*bsize - self.n) # also sets self.n = b * bsize

def download(root, file_name = 'Dataset', unzip = True, OpenFOAM = False):
    """
    Download AirfRANS dataset.

    Args:
        root (str): Root directory where the dataset will be downloaded and unzipped.
        file_name (str, optional): Name of the dataset file. Default: ``'Dataset'``   
        unzip (bool, optional): If ``True``, unzip the dataset file. Default: ``True``
        OpenFOAM (bool, optional): If ``True``, it will download the raw OpenFOAM simulation
            with no post-processing to manipulate it through PyVista. If ``False``, it will 
            download the ``.vtu`` and ``.vtp`` of cropped simulations with a reduced quantity
            of features. Those cropped simulations have been used to train models proposed
            in the associated paper. Default: ``False``
    """
    if OpenFOAM:
        url = 'https://data.isir.upmc.fr/extrality/NeurIPS_2022/OF_dataset.zip'
    else:
        url = 'https://data.isir.upmc.fr/extrality/NeurIPS_2022/Dataset.zip'

    os.makedirs(root, exist_ok = True)
    with DownloadProgressBar(unit = 'B', unit_scale = True, miniters = 1, unit_divisor = 1024, desc = 'Downloading AirfRANS') as t:
        urlretrieve(url, filename = osp.join(root, file_name + '.zip'), reporthook = t.update_to)

    if unzip:
        print("Extracting " + file_name + ".zip at " + root + "...")
        with zipfile.ZipFile(osp.join(root, file_name + '.zip'), 'r') as zipf:
            zipf.extractall(root)

def load(root, task, train = True):
    """
    The different tasks (``'full'``, ``'scarce'``, ``'reynolds'``,
    ``'aoa'``) define the utilized training and test splits. Please note
    that the test set for the ``'full'`` and ``'scarce'`` tasks are the same.
    Each simulation is given as a point cloud defined via the nodes of the
    simulation mesh. Each point of a point cloud is described via 7
    features: its position (in meters), the inlet velocity (two components in meter per second), the
    distance to the airfoil (one component in meter), and the normals (two
    components in meter, set to 0 if the point is not on the airfoil).

    Each point is given a target of 4 components for the underlying regression
    task: the velocity (two components in meter per second), the pressure
    divided by the specific mass (one component in meter squared per second
    squared), the turbulent kinematic viscosity (one component in meter squared
    per second).

    Finally, a boolean is attached to each point to inform if this point lies on
    the airfoil or not.

    The output is a tuple of a list of np.ndarray of shape `(N, 7 + 4 + 1)`, where N is the
    number of points in each simulation and where the features are ordered as presented
    in this documentation, and a list of name for the each corresponding simulation.

    We highly recommend to handle those data with the help of a Geometric Deep
    Learning library such as PyTorch Geometric or Deep Graph Library.

    Args:
        root (string): Root directory where the simulation directories have been saved.
        task (string): The task to study (``'full'``, ``'scarce'``, ``'reynolds'``, ``'aoa'``) 
            that defines the utilized training and test splits.
        train (bool, optional): If ``True``, loads the training dataset, otherwise the 
            test dataset. Default: ``True``
    """
    tasks = ['full', 'scarce', 'reynolds', 'aoa']
    if task not in tasks:
        raise ValueError(f"Expected 'task' to be in {tasks} "
                            f"got '{task}'")

    taskk = 'full' if task == 'scarce' and not train else task
    split = 'train' if train else 'test'

    with open(osp.join(root, 'manifest.json'), 'r') as f:
        manifest = json.load(f)[taskk + '_' + split]

    data_list = []
    name_list = []
    for s in tqdm(manifest, desc = f'Loading dataset (task: {taskk}, split: {split})'):
        simulation = Simulation(root = root, name = s)
        inlet_velocity = (np.array([np.cos(simulation.angle_of_attack),\
                np.sin(simulation.angle_of_attack)])*simulation.inlet_velocity).reshape(1, 2)\
                *np.ones_like(simulation.sdf)

        attribute = np.concatenate([
            simulation.position,
            inlet_velocity,
            simulation.sdf,
            simulation.normals,
            simulation.velocity,
            simulation.pressure,
            simulation.nu_t,
            simulation.surface.reshape(-1, 1)
        ], axis = -1)

        data_list.append(attribute)
        name_list.append(s)
    
    return data_list, name_list



import numpy as np

def cell_sampling_2d(cell_points, cell_attr = None):
    '''
    Sample points in a two dimensional cell via parallelogram sampling and triangle interpolation via barycentric coordinates.
    The vertices have to be ordered in a certain way.

    Args:
        cell_points (np.ndarray): Vertices of the 2 dimensional cells. Shape `(N, 4)` for N cells with 4 vertices.
        cell_attr (np.ndarray, optional): Features of the vertices of the 2 dimensional cells. Shape `(N, 4, k)` for N cells with 4 edges and k features. 
            If given shape `(N, 4)` it will resize it automatically in a `(N, 4, 1)` tensor. Default: ``None``
    '''
    # Sampling via triangulation of the cell and parallelogram sampling
    v0, v1 = cell_points[:, 1] - cell_points[:, 0], cell_points[:, 3] - cell_points[:, 0]
    v2, v3 = cell_points[:, 3] - cell_points[:, 2], cell_points[:, 1] - cell_points[:, 2]  
    a0, a1 = np.abs(np.linalg.det(np.hstack([v0[:, :2], v1[:, :2]]).reshape(-1, 2, 2))), \
                np.abs(np.linalg.det(np.hstack([v2[:, :2], v3[:, :2]]).reshape(-1, 2, 2)))
    p = a0/(a0 + a1)
    index_triangle = np.random.binomial(1, p)[:, None]
    u = np.random.uniform(size = (len(p), 2))
    sampled_point = index_triangle*(u[:, 0:1]*v0 + u[:, 1:2]*v1) + (1 - index_triangle)*(u[:, 0:1]*v2 + u[:, 1:2]*v3)
    sampled_point_mirror = index_triangle*((1 - u[:, 0:1])*v0 + (1 - u[:, 1:2])*v1) + (1 - index_triangle)*((1 - u[:, 0:1])*v2 + (1 - u[:, 1:2])*v3)
    reflex = (u.sum(axis = 1) > 1)
    sampled_point[reflex] = sampled_point_mirror[reflex]

    # Interpolation on a triangle via barycentric coordinates
    if cell_attr is not None:
        t0, t1, t2 = np.zeros_like(v0), index_triangle*v0 + (1 - index_triangle)*v2, index_triangle*v1 + (1 - index_triangle)*v3
        w = (t1[:, 1] - t2[:, 1])*(t0[:, 0] - t2[:, 0]) + (t2[:, 0] - t1[:, 0])*(t0[:, 1] - t2[:, 1])
        w0 = (t1[:, 1] - t2[:, 1])*(sampled_point[:, 0] - t2[:, 0]) + (t2[:, 0] - t1[:, 0])*(sampled_point[:, 1] - t2[:, 1])
        w1 = (t2[:, 1] - t0[:, 1])*(sampled_point[:, 0] - t2[:, 0]) + (t0[:, 0] - t2[:, 0])*(sampled_point[:, 1] - t2[:, 1])
        w0, w1 = w0/w, w1/w
        w2 = 1 - w0 - w1
        
        if len(cell_attr.shape) == 2:
            cell_attr = cell_attr[:, :, None]
        attr0 = index_triangle*cell_attr[:, 0] + (1 - index_triangle)*cell_attr[:, 2]
        attr1 = index_triangle*cell_attr[:, 1] + (1 - index_triangle)*cell_attr[:, 1]
        attr2 = index_triangle*cell_attr[:, 3] + (1 - index_triangle)*cell_attr[:, 3]
        sampled_attr = w0[:, None]*attr0 + w1[:, None]*attr1 + w2[:, None]*attr2

    sampled_point += index_triangle*cell_points[:, 0] + (1 - index_triangle)*cell_points[:, 2]    

    return np.hstack([sampled_point[:, :2], sampled_attr]) if cell_attr is not None else sampled_point[:, :2]

def cell_sampling_1d(line_points, line_attr = None):
    '''
    Sample points in a one dimensional cell via linear sampling and interpolation.

    Args:
        line_points (np.ndarray): Edges of the 1 dimensional cells. Shape `(N, 2)` for N cells with 2 edges.
        line_attr (np.ndarray, optional): Features of the edges of the 1 dimensional cells. Shape `(N, 2, k)` for N cells with 2 edges and k features.
            If given shape `(N, 2)` it will resize it automatically in a `(N, 2, 1)` tensor. Default: ``None``
    '''
    # Linear sampling
    u = np.random.uniform(size = (len(line_points), 1))
    sampled_point = u*line_points[:, 0] + (1 - u)*line_points[:, 1]

    # Linear interpolation
    if line_attr is not None:   
        if len(line_attr.shape) == 2:
            line_attr = line_attr[:, :, None]
        sampled_attr = u*line_attr[:, 0] + (1 - u)*line_attr[:, 1]

    return np.hstack([sampled_point[:, :2], sampled_attr]) if line_attr is not None else sampled_point[:, :2]


import os
import os.path as osp
import numpy as np
import pyvista as pv

from airfrans.reorganize import reorganize
from airfrans.naca_generator import camber_line
import airfrans.sampling as sampling

class Simulation:
    """
    Wrapper to make the study of AirfRANS simulations easier. It only takes in input the location of the
    ``.vtu`` and ``.vtp`` files from the internal and airfoil patches of the simulation.
    """
    def __init__(self, root, name, T = 298.15):
        """
        Define the air properties at temperature ``T``, the parameters, 
        input features and targets for the simulation.

        Args:
            root (str): Root directory where the dataset is.
            name (str): Name of the simulation.
            T (float, optional): Temperature set for the simulation in Kelvin (in incompressible settings, this only defines 
                the air properties. Simulations do not depend on temperature, only on the Reynolds number). Default: 298.15
        """

        self.root = root
        self.name = name
        self.T = np.array(T) # Temperature in Kelvin

        self.reset()
    
    def reset(self):        
        self.MOL = np.array(28.965338e-3) # Air molar weigth in kg/mol
        self.P_ref = np.array(1.01325e5) # Pressure reference in Pa
        self.RHO = self.P_ref*self.MOL/(8.3144621*self.T) # Specific mass of air at temperature T
        self.NU = -3.400747e-6 + 3.452139e-8*self.T + 1.00881778e-10*self.T**2 - 1.363528e-14*self.T**3 # Approximation of the kinematic viscosity of air at temperature T
        self.C = 20.05*np.sqrt(self.T) # Approximation of the sound velocity of air at temperature T        

        self.inlet_velocity = np.array(float(self.name.split('_')[2])).astype('float64')
        self.angle_of_attack = np.array(float(self.name.split('_')[3])*np.pi/180).astype('float64')

        self.internal = pv.read(osp.join(self.root, self.name, self.name + '_internal.vtu'))
        self.airfoil = pv.read(osp.join(self.root, self.name, self.name + '_aerofoil.vtp'))
        self.internal = self.internal.compute_cell_sizes(length = False, volume = False)
        self.airfoil = self.airfoil.compute_cell_sizes(area = False, volume = False)

        # Input candidates
        self.surface = np.array((self.internal.point_data['U'][:, 0] == 0))
        self.sdf = -np.array(self.internal.point_data['implicit_distance'][:, None]).astype('float64')
        self.input_velocity = (np.array([np.cos(self.angle_of_attack), np.sin(self.angle_of_attack)])*self.inlet_velocity).reshape(1, 2)*np.ones_like(self.sdf)
        
        self.position = np.array(self.internal.points[:, :2]).astype('float64')
        self.airfoil_position = np.array(self.airfoil.points[:, :2]).astype('float64')
        self.airfoil_normals = np.array(-self.airfoil.point_data['Normals'][:, :2]).astype('float64')

        self.normals = np.zeros_like(self.input_velocity).astype('float64')
        self.normals[self.surface] = reorganize(
            self.airfoil_position,
            self.position[self.surface], 
            self.airfoil_normals
            )

        # Targets
        self.velocity = np.array(self.internal.point_data['U'][:, :2]).astype('float64')
        self.pressure = np.array(self.internal.point_data['p'][:, None]).astype('float64')
        self.nu_t = np.array(self.internal.point_data['nut'][:, None]).astype('float64')

    def sampling_volume(self, n, density = 'uniform', targets = True):
        """
        Sample points in the internal mesh following the given density.

        Args:
            n (int): Number of sampled points.
            density (str, optional): Density from which the sampling is done. Choose between ``'uniform'`` and ``'mesh_density'``. Default: ``'uniform'``
            targets (bool, optional): If ``True``, velocity, pressure and kinematic turbulent viscosity will be returned in the output ndarray. Default: ``True``
        """
        if density == 'uniform': # Uniform sampling strategy
            p = np.array(self.internal.cell_data['Area']/self.internal.cell_data['Area'].sum())
            sampled_cell_indices = np.random.choice(self.internal.n_cells, size = n, replace = True, p = p)
        elif density == 'mesh_density': # Sample via mesh density
            sampled_cell_indices = np.random.choice(self.internal.n_cells, size = n, replace = True)

        cell_dict = np.array(self.internal.cells).reshape(-1, 5)[sampled_cell_indices, 1:]
        cell_points = self.position[cell_dict]

        if targets:
            cell_attr = np.concatenate([self.sdf[cell_dict], self.velocity[cell_dict], self.pressure[cell_dict], self.nu_t[cell_dict]], axis = -1)
        else:
            cell_attr = self.sdf[cell_dict]            

        return sampling.cell_sampling_2d(cell_points, cell_attr)
    
    def sampling_surface(self, n, density = 'uniform', targets = True):
        """
        Sample points in the airfoil mesh following the given density.

        Args:
            n (int): Number of sampled points.
            density (str, optional): Density from which the sampling is done. Choose between ``'uniform'`` and ``'mesh_density'``. Default: ``'uniform'``
            targets (bool, optional): If ``True``, velocity, pressure and kinematic turbulent viscosity will be returned in the output ndarray. Default: ``True``
        """
        if density == 'uniform': # Uniform sampling strategy
            p = np.array(self.airfoil.cell_data['Length']/self.airfoil.cell_data['Length'].sum())
            sampled_line_indices = np.random.choice(self.airfoil.n_cells, size = n, replace = True, p = p)
        elif density == 'mesh_density': # Sample via mesh density
            sampled_line_indices = np.random.choice(self.airfoil.n_cells, size = n, replace = True)

        line_dict = np.array(self.airfoil.lines).reshape(-1, 3)[sampled_line_indices, 1:]
        line_points = np.array(self.airfoil.points)[line_dict]

        normal = np.array(-self.airfoil.point_data['Normals'][line_dict, :2]) 

        if targets:
            line_attr = np.concatenate([
                normal, 
                np.array(self.airfoil.point_data['U'][line_dict, :2]), 
                np.array(self.airfoil.point_data['p'][line_dict, None]), 
                np.array(self.airfoil.point_data['nut'][line_dict, None])
                ], axis = -1)
        else:
            line_attr = normal     

        return sampling.cell_sampling_1d(line_points, line_attr)

    def sampling_mesh(self, n, targets = True):
        """
        Sample points over the simulation mesh without replacement.

        Args:
            n (int): Number of sampled points, this number has to be lower than the total number of points in the mesh.
            targets (bool, optional): If ``True``, velocity, pressure and kinematic turbulent viscosity will be returned in the output ndarray. Default: ``True``
        """
        idx = np.random.choice(self.position.shape[0], size = n, replace = False)

        # Position
        position = self.position[idx]

        # Inputs
        surface = self.surface[idx]
        sdf = self.sdf[idx]
        normals = self.normals[idx]
        input_velocity = self.input_velocity[idx]

        if targets:
            velocity = self.velocity[idx]
            pressure = self.pressure[idx]
            nu_t = self.nu_t[idx]
            sample = np.concatenate([position, surface[:, None], sdf, normals, input_velocity, velocity, pressure, nu_t], axis = -1)
        else:
            sample = np.concatenate([position, surface[:, None], sdf, normals, input_velocity], axis = -1)

        return sample

    def wallshearstress(self, over_airfoil = False, reference = False):
        """
        Compute the wall shear stress.

        Args:
            over_airfoil (bool, optional): If ``True``, return the wall shear stress over the airfoil mesh. If ``False``,
                return the wall shear stress over the internal mesh. Default: ``False``
            reference (bool, optional): If ``True``, return the wall shear stress computed with the reference velocity field.
                If ``False``, compute the wall shear stress with the velocity attribute of the class. Default: ``False``     
        """
        if reference:
            internal = self.internal.copy()
            internal.point_data['U'] = internal.point_data['U'].astype('float64')
        else:
            internal = self.internal.copy()
            internal.point_data['U'] = np.concatenate([self.velocity, np.zeros_like(self.sdf)], axis = -1)
        jacobian = np.array(internal.compute_derivative(scalars = 'U', gradient = 'jacobian').point_data['jacobian'].reshape(-1, 3, 3)[self.surface, :2, :2]).astype('float64')

        s = .5*(jacobian + jacobian.transpose(0, 2, 1))
        s = s - s.trace(axis1 = -2, axis2 = -1).reshape(-1, 1, 1)*np.eye(2)[None]/3

        wss_ = 2*self.NU.reshape(-1, 1, 1)*s
        wss_ = (wss_*self.normals[self.surface, :2].reshape(-1, 1, 2)).sum(axis = 2)

        if over_airfoil:
            wss = reorganize(self.position[self.surface], self.airfoil_position, wss_)
        else:
            wss = np.zeros_like(self.velocity).astype('float64')
            wss[self.surface] = wss_

        return wss
    
    def force(self, compressible = False, reference = False):
        """
        Compute the force acting on the airfoil. The output is a tuple of the form `(f, fp, fv)`, 
        where `f` is the force, `fp` the pressure contribution of the force and `fv` the viscous
        contribution of the force.

        Args:
            compressible (bool, optional): If ``False``, multiply the force computed with the simulation field by the specific mass. Default: ``False``
            reference (bool, optional): If ``True``, return the force computed with the reference fields.
                If ``False``, compute the force with the fields attribute of the class. Default: ``False``   
        """
        wss = self.wallshearstress(over_airfoil = True, reference = reference)

        if reference:
            p = np.array(self.internal.point_data['p'][:, None]).astype('float64')
        else:
            p = self.pressure
        p = reorganize(self.position[self.surface], self.airfoil_position, p[self.surface])

        airfoil = self.airfoil.copy()
        airfoil.point_data['wallShearStress'] = wss
        airfoil.point_data['p'] = p
        airfoil = airfoil.ptc(pass_point_data = False)

        wp_int = -airfoil.cell_data['p'][:, None]*airfoil.cell_data['Normals'][:, :2]

        wss_int = (airfoil.cell_data['wallShearStress']*airfoil.cell_data['Length'].reshape(-1, 1)).sum(axis = 0)
        wp_int = (wp_int*airfoil.cell_data['Length'].reshape(-1, 1)).sum(axis = 0)

        if compressible:
            force_p = np.array(-wp_int)
            force_v = np.array(wss_int)
        else:
            force_p = np.array(-wp_int)*self.RHO
            force_v = np.array(wss_int)*self.RHO
        force = force_p + force_v

        return force, force_p, force_v

    def force_coefficient(self, reference = False):
        """
        Compute the force coefficients for the simulation. The output is a tuple of the form `((cd, cdp, cdv), (cl, clp, clv))`,
        where `cd` is the drag coefficient, `cdp` the pressure contribution of the drag coefficient and `cdv` the viscous
        contribution of the drag coefficient. Same for the lift coefficient `cl`.

        Args:
            reference (bool, optional): If ``True``, return the force coefficients computed with the reference fields.
                If ``False``, compute the force coefficients with the fields attribute of the class. Default: ``False``   
        """
        f, fp, fv = self.force(reference = reference)

        basis = np.array([[np.cos(self.angle_of_attack), np.sin(self.angle_of_attack)], [-np.sin(self.angle_of_attack), np.cos(self.angle_of_attack)]])
        fp_rot, fv_rot = np.matmul(basis, fp), np.matmul(basis, fv)
        cp, cv = fp_rot/(.5*self.RHO*self.inlet_velocity**2), fv_rot/(.5*self.RHO*self.inlet_velocity**2)
        cdp, cdv = cp[0], cv[0]
        clp, clv = cp[1], cv[1]
        cd, cl = cdp + cdv, clp + clv

        return ((cd, cdp, cdv), (cl, clp, clv))
    
    def mean_absolute_error(self):
        """
        Compute the mean absolute error between the reference target fields and the attribute target fields of the class.
        The target fields are given in this order: velocity_x, velocity_y, pressure, kinematic turbulent viscosity.
        """
        velocity_ref = np.array(self.internal.point_data['U'][:, :2])
        pressure_ref = np.array(self.internal.point_data['p'][:, None])
        nu_t_ref = np.array(self.internal.point_data['nut'][:, None])

        absolute_error = np.abs(np.concatenate([velocity_ref - self.velocity, pressure_ref - self.pressure, nu_t_ref - self.nu_t], axis = -1))

        return absolute_error.mean(axis = 0)

    def mean_squared_error(self):
        """
        Compute the mean squared error between the reference target fields and the attribute target fields of the class.
        The target fields are given in this order: velocity_x, velocity_y, pressure, kinematic turbulent viscosity.
        """
        velocity_ref = np.array(self.internal.point_data['U'][:, :2])
        pressure_ref = np.array(self.internal.point_data['p'][:, None])
        nu_t_ref = np.array(self.internal.point_data['nut'][:, None])

        squared_error = np.concatenate([velocity_ref - self.velocity, pressure_ref - self.pressure, nu_t_ref - self.nu_t], axis = -1)**2

        return squared_error.mean(axis = 0)
    
    def r_squared(self):
        """
        Compute the r_squared between the reference target fields and the attribute target fields of the class.
        The target fields are given in this order: velocity_x, velocity_y, pressure, kinematic turbulent viscosity.
        """
        velocity_ref = np.array(self.internal.point_data['U'][:, :2])
        pressure_ref = np.array(self.internal.point_data['p'][:, None])
        nu_t_ref = np.array(self.internal.point_data['nut'][:, None])
        ref = np.concatenate([velocity_ref, pressure_ref, nu_t_ref], axis = -1)
        pred = np.concatenate([self.velocity, self.pressure, self.nu_t], axis = -1)

        return 1 - ((ref - pred)**2).sum(axis = 0)/((ref.mean(axis = 0) - pred)**2).sum(axis = 0)
    
    def coefficient_relative_error(self):
        """
        Compute the mean relative error between the reference force coefficient and the force coefficient
        computed with the attribute target fields of the class.
        The force coefficients are given in this order: drag coefficient, lift coefficient.
        """
        cd_ref, cl_ref = self.force_coefficient(reference = True)
        cd_pred, cl_pred = self.force_coefficient(reference = False)
        cd_ref, cl_ref = cd_ref[0], cl_ref[0]
        cd_pred, cl_pred = cd_pred[0], cl_pred[0]

        return np.abs((cd_ref - cd_pred)/(cd_ref + 1e-15)), np.abs((cl_ref - cl_pred)/(cl_ref + 1e-15))

    def boundary_layer(self, x, y = 0.1, extrado = True, direction = 'vertical', local_frame = False, resolution = 1000, compressible = False, reference = False):
        """
        Return the boundary layer profile or the trail profile at abscissas x over a line of length y. 
        
        The fields are returned in the following order: the position on the line (in chord length), the
        first component of the velocity in the chosen frame normalized by the inlet velocity, the
        second component of the velocity in the chosen frame normalized by the inlet velocity, the
        pressure normalized by the inlet dynamic pressure, and the turbulent kinematic viscosity
        normalized by the knimatic viscosity.

        Args:
            x (float): Abscissa in chord length. It must be strictly positive. If x < 1, return the boundary layer.
                If x >= 1, return the trail profile.
            y (float, optional): Length of the sampling line. If x < 1, this length is taken from the airfoil surface. If x >= 1,
                this length is taken from one side to the other of the trail. Default: 0.1
            extrado (bool, optional): If ``True``, the boundary layer of the extrado is returned, If ``False``, the boundary
                layer of the intrado is returned. If x>= 1, this parameter is not taken into account. Default: ``True``
            direction (str, optional): Choose between ``'vertical'`` and ``'normals'``. If x < 1, the sampling line is defined 
                as the line starting at the surface of the airfoil at abscissa x, of length y, in the direction of the normals if 
                ``'normals'`` and in the vertical direction if ``'vertical'``. If x>= 1, this parameter is not taken into account
                and the vertical direction is adopted. Default: ``'vertical'``
            local_frame (bool, optional): If ``True``, the sampled velocity components along the lines are given in the local frame,
                i.e. the frame defined by the normals. Else, the sampled velocity components are given in the cartesian frame.
                If x>= 1, this parameter is not taken into account and the cartesian frame is adopted. Default: False
                resolution (int, optional): Resolution of the sampling. Default: 1000
            compressible (bool, optional): If ``True``, add the specific mass to the normalization constant for the pressure.
                Default: ``False``
            reference (bool, optional): If ``True``, return the sampled fields of reference. If ``False``, return the sampled
                fields from the class attribute. Default: ``False``
        """
        assert x > 0, 'x must be stricly positive.'

        if x < 1:
            digits = np.array(list(map(float, self.name.split('_')[4:-1]))).astype('float64')
            camber = camber_line(digits, self.airfoil_position[:, 0])[0]
            idx_extrado = self.airfoil_position[:, 1] > camber

            if extrado:
                arg = np.argmin(np.abs(self.airfoil_position[idx_extrado, 0] - x))
                arg = np.argwhere(idx_extrado.cumsum(axis = 0) == arg)[0][0]
            else:
                arg = np.argmin(np.abs(self.airfoil_position[~idx_extrado, 0] - x))
                arg = np.argwhere((~idx_extrado).cumsum(axis = 0) == arg)[0][0]

            if direction == 'normals':
                normals = np.concatenate([self.airfoil_normals[arg], np.zeros(1)])
            
            elif direction == 'vertical':
                normals = np.array([0, 2*int(extrado) - 1, 0]).astype('float64')
            
            a, b = np.concatenate([self.airfoil_position[arg], .5*np.ones(1)]), np.concatenate([self.airfoil_position[arg], .5*np.ones(1)]) + y*normals
        else:
            c = (x - 1)*np.tan(self.angle_of_attack)
            a, b = np.array([x, c - y/2, self.internal.points[0, 2]]), np.array([x, c + y/2, self.internal.points[0, 2]])

        if reference:
            internal = self.internal.copy()
            internal.point_data['U'] = internal.point_data['U'].astype('float64')
            internal.point_data['p'] = internal.point_data['p'].astype('float64')
            internal.point_data['nut'] = internal.point_data['nut'].astype('float64')
            internal.cell_data['U'] = internal.cell_data['U'].astype('float64')
            internal.cell_data['p'] = internal.cell_data['p'].astype('float64')
            internal.cell_data['nut'] = internal.cell_data['nut'].astype('float64')
        else:
            internal = self.internal.copy()
            internal.point_data['U'] = np.concatenate([self.velocity, np.zeros_like(self.sdf)], axis = -1)
            internal.point_data['p'] = self.pressure.flatten()
            internal.point_data['nut'] = self.nu_t.flatten()
            internal = internal.ptc(pass_point_data = True)

        bl = internal.sample_over_line(a, b, resolution = resolution)
        
        if local_frame and x < 1:
            rot = np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 1]]).astype('float64')
            u = (np.array(bl.point_data['U']).astype('float64')*(np.matmul(rot, normals))).sum(axis = 1)
            v = (np.array(bl.point_data['U']).astype('float64')*normals).sum(axis = 1)
        else:
            u = np.array(bl.point_data['U'][:, 0]).astype('float64')
            v = np.array(bl.point_data['U'][:, 1]).astype('float64')
        
        if compressible:
            p = np.array(bl.point_data['p']).astype('float64')/(.5*self.RHO*self.inlet_velocity**2)
        else:
            p = np.array(bl.point_data['p']).astype('float64')/(.5*self.inlet_velocity**2)
        nut = np.array(bl.point_data['nut']).astype('float64')

        if x < 1:
            yc = np.array(bl.points[:, 1]).astype('float64') - a[1]           
        else:
            yc = np.array(bl.points[:, 1]).astype('float64') - (a[1] + b[1])/2

        if x > 1:
            mask = (u != 0)
            yc, u, v, p, nut = yc[mask], u[mask], v[mask], p[mask], nut[mask]

        return yc, u/self.inlet_velocity, v/self.inlet_velocity, p, nut/self.NU

    def save(self, root):
        """
        Save the internal and the airfoil patches with the attribute targets fields of the class in the root directory.

        Args:
            root (str): Root directory where the files will be saved.
        """    
        internal = self.internal.copy()
        internal.point_data['U'] = np.concatenate([self.velocity, np.zeros_like(self.velocity[:, :1])], axis = -1)
        internal.point_data['p'] = self.pressure.flatten()
        internal.point_data['nut'] = self.nu_t.flatten()
        internal = internal.ptc(pass_point_data = True)

        airfoil = self.airfoil.copy()
        airfoil.point_data['p'] = reorganize(self.position, self.airfoil_position, self.pressure).flatten()
        airfoil.point_data['wallShearStress'] = np.concatenate([self.wallshearstress(over_airfoil = True), np.zeros(airfoil.points[:, :1].shape)], axis = -1)
        airfoil = airfoil.ptc(pass_point_data = True)

        cwd = os.getcwd()
        os.chdir('/')
        os.makedirs(osp.join(root, self.name), exist_ok = True)
        internal.save(osp.join(root, self.name, self.name + '_internal_predicted.vtu'))
        airfoil.save(osp.join(root, self.name, self.name + '_aerofoil_predicted.vtp'))
        os.chdir(cwd)


import numpy as np

def thickness_dist(t, x, CTE = True):
    """
    Standard NACA profile to warp with the help of a camber line to define all the
    4 and 5 digits profiles.

    Args:
        t (float): Thickness of the airfoil in percentage of the chord length.
        x (np.ndarray): Abscissas in chord unit.
        CTE (bool, optional): If ``True`` the profile will be closed at the trailing edge. Default: ``True``
    """
    # CTE for close trailing edge
    if CTE:
        a = -0.1036
    else:
        a = -0.1015
    return 5*t*(0.2969*np.sqrt(x) - 0.1260*x - 0.3516*x**2 + 0.2843*x**3 + a*x**4)

def camber_line(params, x):
    """
    Camber line definition for the NACA 4 and 5 digits series.

    Args:
        params (np.ndarray): Parameters of the NACA 4 or 5 digits profile (ndarray of shape `(3)` or `(4)`).
        x (np.ndarray): Abscissas in chord unit.
    """
    y_c = np.zeros_like(x)
    dy_c = np.zeros_like(x)

    if len(params) == 2:
        m = params[0]/100
        p = params[1]/10

        if p == 0:
            dy_c = -2*m*x
            return y_c, dy_c
        elif p == 1:
            dy_c = 2*m*(1 - x)
            return y_c, dy_c

        mask1 = (x < p)
        mask2 = (x >= p)
        y_c[mask1] = (m/p**2)*(2*p*x[mask1] - x[mask1]**2)
        dy_c[mask1] = (2*m/p**2)*(p - x[mask1])
        y_c[mask2] = (m/(1 - p)**2)*((1 - 2*p) + 2*p*x[mask2] - x[mask2]**2)
        dy_c[mask2] = (2*m/(1 - p)**2)*(p - x[mask2])

    elif len(params) == 3:
        l, p, q = params
        c_l, x_f = 3/20*l, p/20

        f = lambda x: x*(1 - np.sqrt(x/3)) - x_f
        df = lambda x: 1 - 3*np.sqrt(x/3)/2
        old_m = np.array(0.5)
        cond = True
        while cond:
            new_m = np.max(np.array([old_m - f(old_m)/df(old_m), 0]))
            cond = (np.abs(old_m - new_m) > 1e-15)
            old_m = new_m        
        m = old_m
        r = (3*m - 7*m**2 + 8*m**3 - 4*m**4)/np.sqrt(m*(1 - m)) - 3/2*(1 - 2*m)*(np.pi/2 - np.arcsin(1 - 2*m))
        k_1 = c_l/r

        mask1 = (x <= m)
        mask2 = (x > m)
        if q == 0:            
            y_c[mask1] = k_1*((x[mask1]**3 - 3*m*x[mask1]**2 + m**2*(3 - m)*x[mask1]))
            dy_c[mask1] = k_1*(3*x[mask1]**2 - 6*m*x[mask1] + m**2*(3 - m))
            y_c[mask2] = k_1*m**3*(1 - x[mask2])
            dy_c[mask2] = -k_1*m**3*np.ones_like(dy_c[mask2])

        elif q == 1:
            k = (3*(m - x_f)**2 - m**3)/(1 - m)**3
            y_c[mask1] = k_1*((x[mask1] - m)**3 - k*(1 - m)**3*x[mask1] - m**3*x[mask1] + m**3)
            dy_c[mask1] = k_1*(3*(x[mask1] - m)**2 - k*(1 - m)**3 - m**3)
            y_c[mask2] = k_1*(k*(x[mask2] - m)**3 - k*(1 - m)**3*x[mask2] - m**3*x[mask2] + m**3)
            dy_c[mask2] = k_1*(3*k*(x[mask2] - m)**2 - k*(1 - m)**3 - m**3)

        else:
            raise ValueError('Q must be 0 for normal camber or 1 for reflex camber.')

    else:
        raise ValueError('The first input must be a tuple of the 2 or 3 digits that represent the camber line.')   

    return y_c, dy_c

def naca_generator(params, nb_samples = 400, scale = 1, origin = (0, 0), cosine_spacing = True, verbose = True, CTE = True):
    """
    Definition of a complete profile from the NACA 4 and 5 digits series.

    Args:
        params (np.ndarray): Parameters of the NACA 4 or 5 digits profile (ndarray of shape `(3)` or `(4)`).
        nb_samples (int, optional): Number of points to define the profile. Default: 400
        scale (float, optional): Chord length in meters. Default: 1
        origine (tuple, optional): Absolute position of the leading edge. Default: `(0, 0)`
        cosine_spacing (bool, optional): If ``True``, points are sampled via a cosine distance instead of uniformly. Default: ``True``
        verbose (bool, optional): Comments on the generation process. Default: ``True``
        CTE (bool, optional): If ``True`` the profile will be closed at the trailing edge. Default: ``True``
    """
    if len(params) == 3:
        params_c = params[:2]
        t = params[2]/100
        if verbose:
            print(f'Generating naca M = {params_c[0]}, P = {params_c[1]}, XX = {t*100}')
    elif len(params) == 4:
        params_c = params[:3]
        t = params[3]/100
        if verbose:
            print(f'Generating naca L = {params_c[0]}, P = {params_c[1]}, Q = {params_c[2]}, XX = {t*100}')
    else:
        raise ValueError('The first argument must be a tuple of the 4 or 5 digits of the airfoil.')    

    if cosine_spacing:
        beta = np.pi*np.linspace(1, 0, nb_samples + 1)
        x = (1 - np.cos(beta))/2
    else:
        x = np.linspace(1, 0, nb_samples + 1)

    y_c, dy_c = camber_line(params_c, x)
    y_t = thickness_dist(t, x, CTE)
    theta = np.arctan(dy_c)
    x_u = x - y_t*np.sin(theta)
    x_l = x + y_t*np.sin(theta)
    y_u = y_c + y_t*np.cos(theta)
    y_l = y_c - y_t*np.cos(theta)
    x = np.concatenate([x_u, x_l[:-1][::-1]], axis = 0)
    y = np.concatenate([y_u, y_l[:-1][::-1]], axis = 0)
    pos = np.stack([
            x*scale + origin[0],
            y*scale + origin[1]
        ], axis = -1
    )
    pos[0], pos[-1] = np.array([1, 0]), np.array([1, 0])
    return pos


# Configuration file for the Sphinx documentation builder.

# -- Project information

import airfrans
import datetime

project = 'AirfRANS'
author = 'Florent Bonnet'
copyright = f'{datetime.datetime.now().year}, {author}'

version = airfrans.__version__

# -- General configuration

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
]

add_module_names = False
autoclass_content = 'both'
autodoc_member_order = 'bysource'

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}
intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

# -- Options for HTML output

html_theme = 'sphinx_rtd_theme'

# -- Options for EPUB output
epub_show_urls = 'footnote'



__version__ = '0.1.2'

import airfrans.dataset as dataset
import airfrans.naca_generator as naca_generator
import airfrans.sampling as sampling
from airfrans.reorganize import reorganize
from airfrans.simulation import Simulation


import numpy as np

def reorganize(in_order_points, out_order_points, quantity_to_reorder):
    """
    Reorder points coming from different source of data with the same positions.
    
    Args:
        in_order_points (np.ndarray): Positions of the features we want to reorder (shape `(N, 2)`).
        out_order_points (np.ndarray): Positions for the new ordering (shape `(N, 2)`).
        quantity_to_reorder (np.ndarray): Features attached to the points we want to reorder (shape `(N, F)`).
    """
    n = out_order_points.shape[0]
    idx = np.zeros(n)
    for i in range(n):
        cond = (out_order_points[i] == in_order_points)
        cond = cond[:, 0]*cond[:, 1]        
        idx[i] = np.argwhere(cond)[0][0]
    idx = idx.astype('int')

    assert (in_order_points[idx] == out_order_points).all()

    return quantity_to_reorder[idx]